/* 
   neon interop test suite
   Copyright (C) 2001, Joe Orton <joe@manyfish.co.uk>
                                                                     
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
  
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
  
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "config.h"

#include <stdlib.h>

#include <ne_props.h>
#include <ne_uri.h>

#include "interop.h"

#define NS "http://webdav.org/neon/interop/"

static const ne_propname props[] = {
    { "DAV:", "getcontentlength" },
    { "DAV:", "getlastmodified" },
    { "DAV:", "displayname" },
    { "DAV:", "resourcetype" },
    { NS, "foo" },
    { NS, "bar" },
    { NULL }
};

#define ELM_resourcetype (NE_ELM_207_UNUSED + 1)
#define ELM_collection (NE_ELM_207_UNUSED + 4)

static const struct ne_xml_elm complex_elms[] = {
    { "DAV:", "resourcetype", ELM_resourcetype, 0 },
    { "DAV:", "collection", ELM_collection, 0 },
    { NULL }
};

struct private {
    int collection;
};

struct results {
    ne_propfind_handler *ph;
    int result;
    char *error;
};

static void d0_results(void *userdata, const char *uri,
		       const ne_prop_result_set *rset)
{
    struct results *r = userdata;
    
    if (strncmp(uri, "http://", 7) == 0) {
	/* Absolute URI */
	uri = strchr(uri+7, '/');
	if (uri == NULL) {
	    NE_DEBUG(NE_DBG_HTTP, "Invalid URI???");
	    return;
	}
    }

    if (uri_compare(uri, i_path)) {
	warning("response href for wrong resource");
    } else {
	struct private *priv = ne_propset_private(rset);
	if (!priv->collection) {
	    r->result = FAIL;
	    r->error = "Root collection did not define {DAV:}collection";
	    return;
	} else {
	    r->result = 0;
	}
    }
}

static void *create_private(void *userdata, const char *uri)
{
    return ne_calloc(sizeof(struct private));
}

static int d0_validate(void *ud, ne_xml_elmid parent, ne_xml_elmid child)
{
    if ((parent == NE_ELM_prop && child == ELM_resourcetype) ||
	(parent == ELM_resourcetype && child == ELM_collection))
	return NE_XML_VALID;
    else
	return NE_XML_DECLINE;
}

static int endelm(void *userdata, const struct ne_xml_elm *elm, 
		  const char *cdata)
{
    struct results *r = userdata;
    struct private *p = ne_propfind_current_private(r->ph);

    if (elm->id == ELM_collection) {
	p->collection = 1;
    }

    return 0;
}

/* Depth 0 PROPFIND on root collection. */
int propfind_d0(void)
{
    int ret;
    struct results r = {0};

    r.ph = ne_propfind_create(i_session, i_path, NE_DEPTH_ZERO);
    
    ne_propfind_set_private(r.ph, create_private, NULL);

    r.result = FAIL;
    r.error = "No responses returned";

    ne_xml_push_handler(ne_propfind_get_parser(r.ph), complex_elms,
			d0_validate, NULL, endelm, &r);

    ret = ne_propfind_named(r.ph, props, d0_results, &r);

    if (r.result) {
	i_am(r.error);
	return r.result;
    }

    return OK;
}

int propfind_invalid(void)
{
    ne_request *req = ne_request_create(i_session, "PROPFIND", i_path);
    const static char invalid_xml[] = "<foo>";

    ne_set_request_body_buffer(req, invalid_xml, strlen(invalid_xml));

    ne_add_depth_header(req, NE_DEPTH_ZERO);

    ONN("PROPFIND with non-well-formed XML request failed",	
	ne_request_dispatch(req));
    
    if (STATUS(400)) {
	warning("PROPFIND with non-well-formed XML gave %d, should be 400",
		GETSTATUS);
    }

    ne_request_destroy(req);

    return OK;
}


static int prop_ok = 0;
char *prop_uri;

int propinit(void)
{
    CONCAT2(prop_uri, i_path, "prop");
    
    ONREQ(upload_foo("prop"));

    prop_ok = 1;
    
    return OK;
}

#define NP (10)
#define DNP (5)

static ne_proppatch_operation pops[NP + 1];
static ne_propname propnames[NP + 1];
static char *values[NP + 1];

#define PS_VALUE "value goes here"

int propset(void)
{
    int n;
    char tmp[100];

    for (n = 0; n < NP; n++) {
	sprintf(tmp, "prop%d", n);
	propnames[n].nspace = NS;
	propnames[n].name = ne_strdup(tmp);
	pops[n].name = &propnames[n];
	pops[n].type = ne_propset;
	sprintf(tmp, "value%d", n);
	values[n] = ne_strdup(tmp);
	pops[n].value = values[n];
    }

    memset(&pops[n], 0, sizeof(pops[n]));
    memset(&propnames[n], 0, sizeof(propnames[n]));	   
    values[n] = NULL;

    PRECOND(prop_ok);

    prop_ok = 0; /* if the PROPPATCH fails, no point in testing further. */
    ONREQ(ne_proppatch(i_session, prop_uri, pops));
    prop_ok = 1;

    return OK;
}

static void pg_results(void *userdata, const char *uri,
		       const ne_prop_result_set *rset)
{
    struct results *r = userdata;
    const char *value;
    const ne_status *status;
    char tmp[500];
    int n;

    r->result = 0;

    for (n = 0; n < NP; n++) {
	value = ne_propset_value(rset, &propnames[n]);
	status = ne_propset_status(rset, &propnames[n]);

	if (values[n] == NULL) {
	    /* We should have received a 404 for this property. */

	    if (value == NULL) {
		if (status == NULL) {
		    warning("Property %d omitted from results with no status",
			    n);
		} else if (status->code != 404) {
		    warning("Status for missing property %d was not 404", n);
		}
	    } else {
		r->result = FAIL;
		r->error = "Deleted property was present";
	    }
	} else {
	    /* We should have a value for this property. */
	    if (value == NULL) {
		sprintf(tmp, "Property %s was not found.", propnames[n].name);
		r->result = FAIL;
		r->error = ne_strdup(tmp);
	    } else if (strcmp(value, values[n])) {
		sprintf(tmp, "Property %s had value %s, expected %s",
			propnames[n].name, value, values[n]);
		/* Check the value matches. */
		r->result = FAIL;
		r->error = ne_strdup(tmp);
	    }
	}
    }
    
}

int propget(void)
{
    struct results r = {0};

    PRECOND(prop_ok);

    r.result = 1;
    r.error = "No responses returned.";

    ONREQ(ne_simple_propfind(i_session, prop_uri, NE_DEPTH_ZERO,
			     propnames, pg_results, &r));

    if (r.result) {
	i_am(r.error);
	return r.result;
    }	

    return OK;
}

int propmove(void)
{
    char *dest;

    CONCAT2(dest, i_path, "prop2");
    
    ONREQ(ne_move(i_session, 0, prop_uri, dest));

    free(prop_uri);
    prop_uri = dest;

    return OK;
}

int propdeletes(void)
{
    int n;
    
    for (n = 0; n < DNP; n++) {
	pops[n].type = ne_propremove;
	values[n] = NULL;
    }

    ONREQ(ne_proppatch(i_session, prop_uri, pops));

    return OK;
}

int propreplace(void)
{
    int n;

    for (n = DNP; n < NP; n++) {
	char tmp[100];
	sprintf(tmp, "newvalue%d", n);
	pops[n].type = ne_propset;
	pops[n].value = values[n] = ne_strdup(tmp);
    }

    ONREQ(ne_proppatch(i_session, prop_uri, pops));

    return OK;
}

int propcleanup(void)
{
    ne_delete(i_session, prop_uri);
    return OK;
}
